name = "ppa_rspine_buffer"
data = dict()

### Buffer node
data[
    "verilog"
] = """
module ppa_rspine_buffer(gin, gout);

	input gin;
	output gout;

	buffer U2(gout,gin);

endmodule
"""

data[
    "vhdl"
] = """
entity ppa_rspine_buffer is
	port (
		gin : in std_logic;
		gout : out std_logic
	);
end entity;

architecture behavior of ppa_rspine_buffer is
begin

U2: buffer
	port map (
		A => gin,
		Y => gout
	);

end architecture;
"""

data["shape"] = "invtriangle"
data["fillcolor"] = "white"

data["ins"] = [("gin", 1, 1, 0)]
data["outs"] = [("gout", 1)]

data["logic"] = lambda gin: [gin]

data["pd"] = 2
data["le"] = [1, 1]



name = "invis"
data = dict()

### Buffer node
data[
    "verilog"
] = """
module invis(A, Y);

	input A;
	output Y;

	assign Y = A;

endmodule
"""

data[
    "vhdl"
] = """
entity invis is
	port (
		A : in std_logic;
		Y : out std_logic
	);
end entity;

architecture behavior of invis is
begin

	Y <= A;

end architecture;
"""

data["shape"] = "point"
data["fixedsize"] = "shape"
data["width"] = 0
data["height"] = 0
data["fillcolor"] = "black"

data["ins"] = [("A", 1, 1, 0)]
data["outs"] = [("Y", 1)]

data["logic"] = lambda A: [A]

data["pd"] = 0
data["le"] = [0, 0]



import importlib.resources
import pathlib
import shutil

import networkx as nx

from .ExpressionNode import ExpressionNode
from .util import (
    hdl_arch,
    hdl_entity,
    hdl_inst,
    hdl_syntax,
    increment_iname,
    increment_wname,
    merge_mapping_into_cells,
    natural_keys,
    parse_mapping,
    parse_net,
    sub_brackets,
    sub_ports,
    wrap_quotes,
)


class ExpressionGraph(nx.DiGraph):
    """Defines a di-graph of arithmetic expressions

    The graph is defined as follows:
    - Each node is a module that computes an arithmetic expression
    - Each edge is a net connecting two node_data

    Attributes:
        name (string): The name of the graph
        next_net (int): The next net name to be used
        next_block (int): The next block name to be used
        blocks (list): The list of blocks in the graph
        in_ports (list of ((string, int), string)): The list of input ports
        out_ports (list of ((string, int), string)): The list of output ports
        in_extras (list of ((string, int), string)):
            The list of extra input ports, caused by equivalence classes
        out_extra (list of ((string, int), string)):
            The list of extra output ports, caused by equivalence classes
    """

    def __init__(self, name="graph", in_ports=None, out_ports=None):
        """Initializes the graph

        Specifying in_ports and out_ports is optional.
        Doing so implies that the graph is already connected inside of a larger
        structure.
        If they are not specified, the graph will finds its own.

        Args:
            name (string): The name of the graph
            in_ports (list of tuples): The list of input ports
            out_ports (list of tuples): The list of output ports
        """

        # Make sure that in_ports and out_ports are both None or both lists
        if any(x is not None for x in [in_ports, out_ports]) and not all(
            x is not None for x in [in_ports, out_ports]
        ):
            raise ValueError(
                ("in_ports and out_ports" "must both be None or both be lists")
            )

        super().__init__()

        self.name = name

        # Save the input and output ports
        self.in_ports = in_ports
        self.out_ports = out_ports
        self.in_extras = []
        self.out_extras = []

        # Procedurally-generated net names start with "n1"
        self.next_net = 1

        # Procedurally-generated block names start with "block1"
        self.next_block = 0
        self.blocks = [None]

    def add_node(self, node, **attr):
        """Adds a node to the graph

        Args:
            node (ExpressionNode): The name of the node to add
            attr (dict): The attributes of the node to add
        """

        if not isinstance(node, ExpressionNode):
            raise TypeError("Node must be an ExpressionNode")

        # Add GraphViz attributes
        kwargs = node.node_data
        kwargs.update(attr)
        kwargs["shape"] = kwargs.get("shape", "square")
        kwargs["fillcolor"] = kwargs.get("fillcolor", "white")
        kwargs["label"] = kwargs.get("label", "")
        kwargs["style"] = kwargs.get("style", "filled")
        kwargs["pos"] = "{0},{1}!".format(node.x_pos * -1, node.y_pos * -1)

        kwargs["verilog"] = wrap_quotes(kwargs.get("verilog", ""))
        kwargs["vhdl"] = wrap_quotes(kwargs.get("vhdl", ""))

        # Add the node to the graph
        node.graph = self
        super().add_node(node, **kwargs)
        return node

    def remove_node(self, node):
        """Removes a node from the graph

        Args:
            node (ExpressionNode): The name of the node to remove
        """

        if not isinstance(node, ExpressionNode):
            raise TypeError("Node must be an ExpressionNode")

        # Remove the node from the graph
        super().remove_node(node)
        return node

    def add_edge(self, parent, pin1, child, pin2):
        """Adds a directed edge to the graph, from parent to child

        Args:
            parent (ExpressionNode): The name of the first node
            pin1 (tuple): (name, index) of the pin on the first node
            child (ExpressionNode): The name of the second node
            pin2 (tuple): (name, index) of the pin on the second node
        """

        if not isinstance(parent, ExpressionNode):
            raise TypeError("Node1 must be an ExpressionNode")
        if not isinstance(child, ExpressionNode):
            raise TypeError("Node2 must be an ExpressionNode")

        # Connect the nodes
        proposed_net_name = "$n{0}_{1}".format(self.next_net, self.name)
        net_name = parent.add_child(child, pin1, pin2, proposed_net_name)
        if proposed_net_name == net_name:
            self.next_net += 1

        # Styles the edge for GraphViz visualization
        kwargs = {
            "arrowhead": "none",
            "ins": [pin1],
            "outs": [pin2],
            "edge_nets": [net_name],
        }

        # Initialize weight to parasitc delay
        # This is later modified by logical effort and cross-track cap
        kwargs["fanout"] = 1
        kwargs["tracks"] = 0
        kwargs["delay"] = child.node_data["pd"]
        kwargs["weight"] = kwargs["delay"]

        # If the two nodes are already connnected, simply update the pins
        if self.has_edge(parent, child):
            edge_data = self.get_edge_data(parent, child, default=kwargs)
            edge_data["ins"].append(pin1)
            edge_data["outs"].append(pin2)
            edge_data["edge_nets"].append(net_name)
        else:
            # Add the edge to the graph
            super().add_edge(parent, child, **kwargs)

        return self.get_edge_data(parent, child)

    def remove_edge(self, parent, child):
        """Removes an edge from the graph

        Args:
            parent (ExpressionNode): The parent node
            child (ExpressionNode): The child node
        """

        if not isinstance(parent, ExpressionNode):
            raise TypeError("Node1 must be an ExpressionNode")
        if not isinstance(child, ExpressionNode):
            raise TypeError("Node2 must be an ExpressionNode")
        if not self.has_edge(parent, child):
            raise ValueError("Edge does not exist")

        # Save the edge data
        edge_data = self.get_edge_data(parent, child)

        # Remove the edge from the graph
        super().remove_edge(parent, child)

        # Remove the edge from the nodes
        parent.remove_child(child)

        # Return the edge data
        return edge_data

    ### NOTE: Improve the heuristic used herein
    def update_edge_weight(self, parent, child, weight_fun=None):
        """Updates the weight of an edge from parent to child

        Args:
            parent (ExpressionNode): The parent node
            child (ExpressionNode): The child node
        """

        if not isinstance(parent, ExpressionNode):
            raise TypeError("Node1 must be an ExpressionNode")
        if not isinstance(child, ExpressionNode):
            raise TypeError("Node2 must be an ExpressionNode")
        if not self.has_edge(parent, child):
            raise ValueError("Edge does not exist")

        ### This is a bad estimate of delay
        if weight_fun is None:
            edge_data = self.get_edge_data(parent, child)
            weight = edge_data["delay"]
            weight += edge_data["fanout"]
            weight += edge_data["tracks"]
        else:
            weight = weight_fun(parent, child)

        edge_data["weight"] = weight
        return weight

    def critical_path(self):
        """Returns the longest path of the graph"""

        # Get valid nodes
        valid_nodes = [
            node
            for node in self.nodes
            if node.block is None and node.equiv_class.rep is node
        ]

        if len(valid_nodes) < 2:
            return []

        # Get subgraph view with only valid nodes
        subgraph = self.subgraph(valid_nodes)

        # Find the critical path
        return nx.dag_longest_path(subgraph)

    def add_block(self, *nodes, graph_type=None):
        """Creates a new module block of nodes

        This is a sub-graph of nodes that will be flattened together into a
        single monolithic module.

        Args:
            nodes (list): The list of nodes to add to the block
        """
        if not all([isinstance(node, ExpressionNode) for node in nodes]):
            raise TypeError("Nodes must be ExpressionNodes")
        if not all([node.block is None for node in nodes]):
            raise ValueError("Nodes must not already be in a block")

        # If the block is empty, don't create it
        if not nodes:
            return None

        # Get the name of the block
        block_id = self.next_block

        # Assign the block to the nodes
        for node in nodes:
            node.block = block_id

        # Create the block
        if graph_type is None:
            new_block = ExpressionGraph(name="block_{0}".format(block_id))
        else:
            new_block = graph_type(name="block_{0}".format(block_id))
        subgraph = self.subgraph(nodes)
        del subgraph.nodes
        new_block.add_nodes_from(subgraph.nodes(data=True))
        new_block.add_edges_from(subgraph.edges(data=True))
        self.blocks[block_id] = new_block
        if block_id == len(self.blocks) - 1:
            self.blocks.append(None)

        # Increment the next block name
        self.next_block = next(
            x for x in range(len(self.blocks)) if self.blocks[x] is None
        )

        return block_id

    def remove_block(self, block_id):
        """Removes a block from the graph

        Args:
            block_id (int): The ID of the block to remove
        """
        if (
            block_id not in range(len(self.blocks))
            or self.blocks[block_id] is None
        ):
            raise ValueError("Invalid block ID")

        # Remove the block from the graph
        block = self.blocks[block_id]
        for node in block:
            node.block = None
        self.blocks[block_id] = None

    def reset_blocks(self):
        """Removes all blocks from the graph"""

        # Remove all blocks from the graph
        for block_id in [x for x in self.blocks if x is not None]:
            self.remove_block(block_id)

        # Reset the next block name
        self.next_block = 0
        self.blocks = [None]

    def add_best_blocks(self, graph_type=None):
        """Groups nodes into blocks based on critical paths"""

        path = self.critical_path()

        # Add a block
        if len(path) > 1:
            self.add_block(*path, graph_type=graph_type)
            # Recurse
            return self.add_best_blocks(graph_type=graph_type)
        return

    def _get_internal_nets(self, null_flag=False):
        """Returns the internal nets of the graph"""

        # Compatibility issue
        if null_flag:
            return (set(), set())

        # Get all nets in the graph
        in_nets = set()
        out_nets = set()

        # Get the nets from nodes
        for node in self.nodes:
            # Ignore any nodes that are not the representative of their class
            if node.equiv_class.rep is not node:
                continue
            for net in node.in_nets.values():
                in_nets.update([parse_net(x) for x in net])
            for net in node.out_nets.values():
                out_nets.update([parse_net(x) for x in net])

        # Get the nets from blocks
        for block in [x for x in self.blocks if x is not None]:
            block_in_nets, block_out_nets = block._get_internal_nets()
            in_nets.update(block_in_nets)
            out_nets.update(block_out_nets)

        # Filter out the graph's ports, as these are not internal wires
        if self.in_ports is not None:
            in_ports = [x[0][0] for x in self.in_ports]
        else:
            in_ports = []
        if self.out_ports is not None:
            out_ports = [x[0][0] for x in self.out_ports]
        else:
            out_ports = []
        in_extras = [x[0][0] for x in self.in_extras]
        out_extras = [x[0][0] for x in self.out_extras]
        all_ports = in_ports + out_ports + in_extras + out_extras
        in_nets = [x for x in in_nets if x.split("[")[0] not in all_ports]
        out_nets = [x for x in out_nets if x.split("[")[0] not in all_ports]

        return (set(in_nets), set(out_nets))

    def _get_ports(self):
        """Returns the ports of the graph"""

        # If the ports are defined, return them
        if self.in_ports is not None:
            return (
                self.in_ports + self.in_extras,
                self.out_ports + self.out_extras,
            )

        # Otherwise, find them from nodes and blocks
        in_ports, out_ports = self._get_internal_nets()

        # Signals generated inside this graph are not inputs
        in_ports = in_ports - out_ports

        # Form the ports
        # Assume that all these retrieved nets are 1-bit
        in_internal = [(sub_brackets(x), 1) for x in in_ports]
        in_external = list(in_ports)
        in_ports = [(x, y) for x, y in zip(in_internal, in_external)]

        out_internal = [(sub_brackets(x), 1) for x in out_ports]
        out_external = list(out_ports)
        out_ports = [(x, y) for x, y in zip(out_internal, out_external)]

        return (in_ports + self.in_extras, out_ports + self.out_extras)

    # NOTE: This function fails flake8 C901
    # TO-DO: Make this function pass flake8 C901
    def hdl(
        self,
        out=None,
        mapping="behavioral",
        language="verilog",
        flat=False,
        block_flat=False,
        node_flat=True,
        merge_mapping=True,
        module_name=None,
        description_string="start of unnamed graph",
    ):
        """Creates a HDL description of the graph

        Args:
            out (str): The file to write the HDL to
            mapping (str): The cell mapping to use for the HDL generation
            language (str): The language in which to generate the HDL
            flat (bool): If True, flatten the graph's HDL
            block_flat (bool): If True, flatten all blocks in the graph's HDL
            node_flat (bool): If True, flatten all cells in the graph's HDL
            merge_mapping (bool): If True, merge the mapping file in
            module_name (str): The name of the module to generate
            description_string (str): String commend to prepend to the HDL

        Returns:
            str: HDL module definition representing the graph
            list: Set of HDL module definitions used in the graph

        """
        # Check that the language is supported
        if language not in ["verilog", "vhdl"]:
            raise ValueError("Unsupported hardware-descriptive language")

        # Set language-specific syntax
        syntax = hdl_syntax[language]

        # If module name is not defined, set it to graph's name
        if module_name is None:
            module_name = self.name

        # Create the HDL

        hdl = ""
        module_defs = set()

        # Pull in the HDL description of blocks
        # This HDL description will have multiple cell-internal wires
        # By default, these are named w*
        # These names need to be made unique
        w_ctr = 0
        for block_id in range(len(self.blocks)):
            if self.blocks[block_id] is None:
                continue
            block = self.blocks[block_id]

            block_hdl, block_defs = block.hdl(
                mapping=mapping,
                language=language,
                flat=block_flat,
                node_flat=node_flat,
                merge_mapping=merge_mapping,
                module_name="{0}_block_{1}".format(module_name, block_id),
                description_string="block {0}".format(block_id),
            )
            block_hdl, w_ctr = increment_wname(block_hdl, w_ctr)
            hdl += block_hdl
            hdl += "\n"
            module_defs.update(block_defs)

        # If the mapping file is intended to be merged in, do so
        if merge_mapping:
            # Parse the mapping file
            file_suffix = syntax["file_extension"]
            map_file = "{0}_map{1}".format(mapping, file_suffix)
            map_path = importlib.resources.path("pptrees.mappings", map_file)
            with map_path as pkg_map_file:
                mapping_dict = parse_mapping(pkg_map_file)

        # Pull in the HDL description of nodes outside of blocks
        # If fully flattening them, need to rename "w*" internal wires
        if node_flat:
            w_ctr = 0
        # Reset the list of extra nets caused by equivalence classes
        self.in_extras = []
        self.out_extras = []
        for node in self:
            # If the node is part of an equivalence class,
            # and some part of the equivalence class is not in this graph,
            # and the node is not part of a bigger subtree,
            # bring up its internal wires to the top level
            if (
                any([x not in self for x in node.equiv_class])
                and node.parent.equiv_class.rep is node.parent
            ):

                # If this node is the representative, its wires become outputs
                if node.equiv_class.rep is node:
                    self.out_extras += [
                        parse_net(wire)
                        for net in node.equiv_class.out_nets.values()
                        for wire in net
                    ]
                # Otherwise, its wires become inputs
                else:
                    self.in_extras += [
                        parse_net(wire)
                        for net in node.equiv_class.out_nets.values()
                        for wire in net
                    ]

            # Check whether the node is in a block, and whether this is it
            # If the node is not inside a block, it's automatically usable
            usable = node.block is None
            # Otherwise, check whether the node is inside THIS block
            if not usable:
                # Query the node for its graph and block ID
                node_block = node.graph.blocks[node.block]
                # Check whether the node's block is this graph
                usable = node_block is self
            if not usable:
                continue
            # Get the node's HDL description
            node_hdl, node_defs = node.hdl(language=language, flat=node_flat)
            # If the cell is flat, rename the internal wires
            if node_flat:
                node_hdl, w_ctr = increment_wname(node_hdl, w_ctr)

            # If the mapping file is intended to be merged in, do so
            if merge_mapping:
                node_hdl = merge_mapping_into_cells(node_hdl, mapping_dict)
                new_defs = set()
                for a in node_defs:
                    new_def = merge_mapping_into_cells(a, mapping_dict)
                    new_def = increment_iname(new_def)
                    new_defs.add(new_def)
                node_defs = new_defs

            hdl += node_hdl
            module_defs.update(node_defs)
        # Format the list of extra nets caused by equivalence classes
        self.in_extras = [((sub_brackets(x), 1), x) for x in self.in_extras]
        self.out_extras = [((sub_brackets(x), 1), x) for x in self.out_extras]

        # This HDL description will have multiple instances in it
        # By default, util.hdl_inst names all instances "U0"
        # These names need to be made unique
        hdl = increment_iname(hdl)

        ### NOTE: Is this general? Improvements wanted
        ### In general, hard-coding an is_block flag sounds like a bad idea
        is_block = list(self.nodes)[0].graph is not self

        # If the graph is a block, then it has individual ports
        # for each bit of every input and output port.
        # So the HDL should reflect through judicious use of net name fixing
        if is_block and not flat:
            (in_ports, out_ports) = self._get_ports()
            for a in in_ports + out_ports:
                hdl = hdl.replace(a[1], a[0][0])

        # Add wire definitions
        # But if this graph is a block,
        # all wires that are inputs into the module are not internal
        # all wires that are outputs from from the cells are not internal
        # therefore, no wires are internal
        in_wires, out_wires = self._get_internal_nets(null_flag=is_block)
        wires = in_wires | out_wires
        wires = sorted(list(wires), key=natural_keys)
        if len(wires) > 0:
            wire_hdl = syntax["wire_def"].format(", ".join(wires))
        else:
            wire_hdl = ""

        # Assemble the HDL
        hdl = (
            syntax["comment_string"]
            + description_string
            + "\n"
            + "\t"
            + wire_hdl
            + "\n"
            + hdl
        )

        # If flat HDL is desired, it can returned here
        if flat:
            (in_ports, out_ports) = self._get_ports()
            hdl = sub_ports(hdl, in_ports + out_ports)
            return hdl, module_defs

        # Otherwise, return the HDL module definition
        hdl, module_defs, file_out_hdl = self._wrap_hdl(
            hdl, module_defs, language, module_name
        )

        # Write the HDL to file
        if out is not None:
            self._write_hdl(file_out_hdl, out, language, mapping, merge_mapping)

        return hdl, module_defs

    def _wrap_hdl(self, hdl, module_defs, language="verilog", module_name=None):
        """Wraps the HDL in a module definition"""

        # If flat HDL is not desired, wrap the graph in a module
        ## First get in_ports and out_ports
        (in_ports, out_ports) = self._get_ports()
        entity_ins = [x[0] for x in in_ports]
        entity_outs = [x[0] for x in out_ports]
        ## Then create the entity
        entity = hdl_entity(module_name, entity_ins, entity_outs, language)
        ## Then create the architecture
        arch = hdl_arch(module_name, hdl, language)
        ## Add the entity and architecture to the module_defs
        formatted_defs = sorted(list(module_defs), key=natural_keys)
        file_out_hdl = entity + arch + "".join(formatted_defs)
        module_defs.add(entity + arch)
        ## Create an instance of the module
        inst_ports = [[x[0][0], x[1]] for x in in_ports + out_ports]
        inst = hdl_inst(module_name, inst_ports, language)
        ## Add the instance to the HDL
        hdl = inst

        return hdl, module_defs, file_out_hdl

    def _write_hdl(
        self,
        file_out_hdl,
        out=None,
        language="verilog",
        mapping="behavioral",
        merge_mapping=True,
    ):
        """Writes the HDL to a file"""
        # Check that output path is valid
        if out is None:
            raise ValueError("Output path must be defined")

        # Set language-specific syntax
        syntax = hdl_syntax[language]

        outdir = pathlib.Path(out).resolve().parent
        if not outdir.exists():
            raise ValueError("Output path does not exist")

        # Verify that the mapping is supported
        file_suffix = syntax["file_extension"]
        map_file = "{0}_map{1}".format(mapping, file_suffix)
        map_path = importlib.resources.path("pptrees.mappings", map_file)
        with map_path as pkg_map_file:
            local_map_file = outdir / (mapping + "_map" + file_suffix)

            if not pkg_map_file.is_file():
                raise ValueError("Unsupported mapping requested")

            if not merge_mapping:
                # Copy the mapping file to the output directory
                shutil.copy(pkg_map_file, local_map_file)

        with open(out, "w") as f:
            f.write(file_out_hdl)


if __name__ == "__main__":
    raise RuntimeError("This module is not intended to be run directly")



name = "ppa_black_ling"
data = dict()

### Black cell
data[
    "verilog"
] = """
module ppa_black_ling(gin, pin, gout, pout);

	input [1:0] gin;
	input pin;
	output gout, pout;

	assign pout = pin;
	or2 U2(gout,gin[0],gin[1]);

endmodule
"""

data[
    "vhdl"
] = """
entity ppa_black_ling is
	port (
		gin : in std_logic_vector(1 downto 0);
		gout : out std_logic;
		pin : in std_logic;
		pout : out std_logic
	);
end entity;

architecture behavior of ppa_black_ling is
begin

	pout <= pin;

U2: or2
	port map (
		A => gin(0),
		B => gin(1),
		Y => gout
	);

end architecture;
"""

# Graphical representation
data["shape"] = "square"
data["fillcolor"] = "black"

# List of inputs represented by (name, total_bits, *bits_per_direction) tuple
data["ins"] = [("gin", 2, 1, 1), ("pin", 1, 1, 0)]

# List of outputs represented by (name, bits) tuple
data["outs"] = [("gout", 1), ("pout", 1)]

# Returns function that executes the logic of the module
data["logic"] = lambda gin, pin: [gin[1] | (pin[1] & gin[0]), pin[1] & pin[0]]

# Logical effort
data["pd"] = 6.0 / 3
data["le"] = [4.0 / 3, 4.0 / 3]



name = "ppa_rspine"
data = dict()

### Grey cell
data[
    "verilog"
] = """
module ppa_rspine(gin, pin, gout);

	input[1:0] gin;
	input pin;
	output gout;

	ao21 U1(gout,gin[0],pin,gin[1]);

endmodule
"""

data[
    "vhdl"
] = """
entity ppa_rspine is
	port (
		gin : in std_logic_vector(1 downto 0);
		gout : out std_logic;
		pin : in std_logic
	);
end entity;

architecture behavior of ppa_rspine is
begin

U1: ao21
	port map (
		A0 => gin(0),
		A1 => pin,
		B0 => gin(1),
		Y => gout
	);

end architecture;
"""

# Graphical representation
data["shape"] = "square"
data["fillcolor"] = "grey"

# List of inputs represented by (name, bits, diagonal_bits) tuple
data["ins"] = [("gin", 2, 1, 1), ("pin", 1, 1, 0)]

# List of outputs represented by (name, bits) tuple
data["outs"] = [("gout", 1)]

# Returns function that executes the logic of the module
data["logic"] = lambda gin, pin: [gin[1] | (pin & gin[0])]

# Logical effort
data["pd"] = 7.5 / 3
data["le"] = [5.0 / 3, 6.0 / 3]



name = "ppa_rspine_pre"
data = dict()

### Pre-processing node for input sparsity
data[
    "verilog"
] = """
module ppa_rspine_pre(a_in, b_in, gout);

	input a_in, b_in;
	output gout;

	and2 U2(gout,a_in,b_in);

endmodule
"""

data[
    "vhdl"
] = """
entity ppa_rspine_pre is
	port (
		a_in : in std_logic;
		b_in : in std_logic;
		gout : out std_logic
	);
end entity;

architecture behavior of ppa_rspine_pre is
begin

U2: and2
	port map (
		A => a_in,
		B => b_in,
		Y => gout
	);

end architecture;
"""

data["shape"] = "square"
data["fillcolor"] = "white"
data["label"] = "pre"
data["style"] = "dashed"

data["ins"] = [("a_in", 1, 1, 0), ("b_in", 1, 1, 0)]
data["outs"] = [("gout", 1)]

data["logic"] = lambda a, b: [0, a & b]

data["pd"] = 2
data["le"] = [4.0 / 3, 4.0 / 3]



name = "ppa_post_plusone"
data = dict()

### Post-processing node
data[
    "verilog"
] = """
module ppa_post_plusone(xin, yin, pin, gin, sum, sum_plus_one);

	input xin, yin, pin, gin;
	output sum, sum_plus_one;

        wire w1;

        or2  U1(w1, pin, gin);
	mux2 U2(sum,gin,xin,yin);
	mux2 U3(sum_plus_one,w1 ,xin,yin);

endmodule
"""

data[
    "vhdl"
] = """
"""

data["shape"] = "circle"
data["label"] = "x"
data["fontname"] = "Comic Sans"
data["style"] = "bold"
data["fixedsize"] = "shape"
data["penwidth"] = "4.0"
data["fontsize"] = "52"

data["ins"] = [("gin", 1, 0, 1), ("xin", 1, 1, 0), ("yin", 1, 1, 0)]
data["outs"] = [("sum", 1), ("sum_plus_one", 1)]

# data["logic"] = lambda yin, xin, gin, pin: [yin if gin else xin]

data["pd"] = 9 / 3
data["le"] = [9 / 3, 9 / 3]



import os
import re
import uuid

import PIL.Image


def lg(x):
    """Returns the base-2 logarithm of x, rounded down"""
    return x.bit_length() - 1


def sub_brackets(x):
    """Reformats 'a[0]' to 'a_0'"""
    return x.replace("[", "_").replace("]", "")


def wrap_quotes(original: str):
    """Wrap double quotes around a string"""
    out = original
    if len(original) < 2 or original[0] != '"' or original[-1] != '"':
        out = f'"{out}"'
    return out


def verso_pin(x):
    """Returns the verso version of a pin"""
    return x.replace("out", "in") if "out" in x else x.replace("in", "out")


def catalan(n):
    """Returns the nth Catalan number"""
    if n == 0:
        return 1
    return catalan(n - 1) * (4 * n - 2) // (n + 1)


def catalan_mirror_point(n):
    """Returns the first safe mirror mark for the nth Catalan number"""
    # Get the halfway mark
    half = catalan(n) >> 1
    # If n is odd, there's no clean split
    if n & 1:
        return half + (catalan(n // 2) ** 2) / 2
    # If n is even, there's a clean split
    else:
        return half


def catalan_bounds(n):
    """Returns the first n catalan numbers, minus one"""
    return [catalan(x) - 1 for x in range(n)]


def match_nodes(parent, child, index):
    """Attempts to match the ports of two nodes

    Args:
        parent (Node): The parent node
        child (Node): The child node
        index (int): The index of the parent's input port
    """
    # If no match is found, will return None
    # Otherwise will return a list of pins to be connected
    # List items will be of the format (parent_pin, child_pin)
    ret = []

    # Get the parent and child ports
    parent_ports = parent.node_data["ins"]
    child_ports = child.node_data["outs"]

    # Iterate over all input ports
    for port in parent_ports:
        if port[2 + index] == 0:
            continue
        try:
            # Get the matching output port
            matching_port = get_matching_port(port, child_ports)
        except ValueError:
            return None
        # Figure out which bits of the port to connect
        offset = 0
        for x in range(index):
            offset += port[2 + x]
        # Add the matching pin pairs to the list
        for b in range(port[2 + index]):
            pin1 = (port[0], offset + b)
            pin2 = (matching_port[0], b)
            ret.append((pin1, pin2))

    return ret


def change_in_nets(node, old_nets, new_nets, index):
    """Overwrites the input nets of a node's parent with new nets

    Args:
        node (Node): The node whose in_nets will change
        old_nets (dict): A dictionary that is node.child.out_nets
        new_nets (dict): A dictionary similar to old_nets
    """
    # Loop through all output nets, making a dictionary
    dic = {}
    for k in new_nets:
        new_port = new_nets[k]
        old_port = old_nets[k]
        verso = verso_pin(k)
        # Check if the net is in the node's inputs
        try:
            node_port = node.in_nets[verso]
        except KeyError:
            continue
        # Check if the net is on the correct side of the node's inputs
        flag = False
        for tup in node.node_data["ins"]:
            if tup[0] == verso and tup[2 + index] == 0:
                flag = True
                break
        if flag:
            continue
        # Change the net names
        for a in range(len(new_port)):
            dic[old_port[a]] = new_port[a]
        node_port = [dic.get(x, x) for x in node_port]
        node.in_nets[verso] = node_port
    return node


def get_matching_port(port, other_ports):
    """Returns the port in other_ports that is the verso of port

    Args:
        port (str): The port to find the verso of
        other_ports (list): The list of ports to search
    """
    for other_port in other_ports:
        if port[0] == verso_pin(other_port[0]):
            return other_port
    raise ValueError("No matching port found for {0}".format(port[0]))


def parse_net(x):
    """Converts a net's ID to its name in HDL

    These come in 3 possible flavors:
        - None (unassigned net) -> parsed to n0
        - Integer (assigned net) -> parsed to n`Integer
        - Hardcoded name ($net_name) -> parsed to net_name
    """
    if x is None:
        return "n0"
    if isinstance(x, int):
        return "n" + str(x)
    if "$" in x:
        return x.replace("$", "")
    raise TypeError("net stored in node {0} is invalid".format(repr(x)))


hdl_syntax = {
    "verilog": {
        "entity": "\nmodule {0}(\n\n\t{1}\n\t);\n\n",
        "entity_in": "input",
        "entity_out": "output",
        "entity_port": "{0} {2} {1},",
        "port_range": "[{0}:{1}]",
        "arch": "{1}\nendmodule // {0}\n",
        "inst": "\n\t{0} U0(\n{1}\n\t);",
        "inst_port": "\t\t.{0}({1})",
        "slice_markers": lambda x: x,
        "wire_def": "wire {0};",
        "comment_string": "// ",
        "file_extension": ".v",
    },
    "vhdl": {
        "entity": "entity {0} is\n\tport (\n{1}\n);\nend entity;\n\n",
        "entity_in": "in",
        "entity_out": "out",
        "entity_port": "\t\t{1} : {0} std_logic{2};\n",
        "port_range": "_vector({0} downto {1})",
        "arch": (
            "architecture {0}_arch of {0} is"
            "\n\tbegin\n{1}\n"
            "end architecture {0}_arch;\n"
        ),
        "inst": ("\tU0: {0}\n" "\t\tport map (\n{1}\n" "\t\t);"),
        "inst_port": "\t\t{0} => {1}\n",
        "slice_markers": lambda x: x.replace("[", "(")
        .replace("]", ")")
        .replace(":", " downto "),
        "wire_def": "signal {0} : std_logic;",
        "comment_string": "-- ",
        "file_extension": ".vhd",
    },
}


def hdl_entity(name, ins, outs, language="verilog"):
    """Formats an entity declaration

    Args:
        name (str): The name of the entity
        ins (list of (string, int)): A list of input ports
        outs (list of (string, int)): A list of output ports
        language (str): The language in which to generate the HDL
    """
    syntax = hdl_syntax[language]
    ports_str = []
    for port in ins:
        port_range = (
            syntax["port_range"].format(port[1] - 1, 0) if port[1] > 1 else ""
        )
        next_port = syntax["entity_port"].format(
            syntax["entity_in"], port[0], port_range
        )
        ports_str.append(next_port)
    for port in outs:
        port_range = (
            syntax["port_range"].format(port[1] - 1, 0) if port[1] > 1 else ""
        )
        next_port = syntax["entity_port"].format(
            syntax["entity_out"], port[0], port_range
        )
        ports_str.append(next_port)
    ports_str = "\n\t".join(ports_str)[:-1]
    return syntax["entity"].format(name, ports_str)


def hdl_arch(name, body, language="verilog"):
    """Formats an architecture declaration

    Args:
        name (str): The name of the entity
        body (str): The body of the architecture
        language (str): The language in which to generate the HDL
    """
    syntax = hdl_syntax[language]
    return syntax["arch"].format(name, body)


def hdl_inst(name, ports, language="verilog"):
    """Formats an instance declaration

    Args:
        name (str): The name of the instance
        ports (list of (string, string)): A list of ports to be connected
        language (str): The language in which to generate the HDL
    """
    syntax = hdl_syntax[language]
    ports = [(port[0], syntax["slice_markers"](port[1])) for port in ports]
    ports_list = [
        syntax["inst_port"].format(port[0], port[1]) for port in ports
    ]
    return syntax["inst"].format(name, ",\n".join(ports_list))


def sub_ports(hdl, ports):
    """Substitutes port in an HDL string with corresponding module ports"""
    for port in sorted(ports, key=lambda x: natural_keys(x[1]), reverse=True):
        ((local, num), remote) = port
        if num == 1:
            hdl = hdl.replace(local, remote)
        ### NOTE: This is a complete 2 AM hack
        ### It assumes that all ports are zero-indexed, forever
        else:
            remote = remote.split("[")[0]
            hdl = hdl.replace(local, remote)
    return hdl


def atoi(x):
    """Converts a string to an integer"""
    return int(x) if x.isdigit() else x


def natural_keys(text):
    """Human sorting / natural sorting"""
    return [atoi(c) for c in re.split(r"(\d+)", text)]


### NOTE: This needs to be reworked when map files are reworked
### THIS IS A HACK
### Currently it assumes that all lines of map files with logic will contain
### either "assign" or "));"
### And that all lines of map files with ports will contain "Y, "
def parse_mapping(mapping_file):
    """Parses a mapping file and returns a dictionary of mappings"""
    mapping = {}
    with open(mapping_file, "r") as f:
        in_a_module = False
        current_module = None
        for line in f:
            split_line = line.strip().split()
            if len(split_line) == 0:
                continue
            if split_line[0] == "module":
                in_a_module = True
                current_module = split_line[1]
                mapping[current_module] = [[], ""]
            elif split_line[0] == "endmodule":
                in_a_module = False
                current_module = None
            elif in_a_module and "Y, " in line:
                new_split = [x.strip() for x in line.strip().split(",")]
                mapping[current_module][0] = new_split
            elif in_a_module and ("));" in line or "assign" in line):
                mapping[current_module][1] += line
    return mapping


### NOTE: This needs to be reworked when map files are reworked
def merge_mapping_into_cells(hdl, mapping):
    """Merges the definitions found inside a mapping into HDL cells"""
    new_hdl = []
    for line in hdl.split("\n"):
        first_word = None
        split_line = line.strip().split()
        if len(split_line) > 0:
            first_word = split_line[0]
        if first_word in mapping:
            net_string = re.search(r"\((.*?)\)", line).group(1)
            nets = net_string.split(",")
            data = mapping[first_word][1]
            behav = "assign" in data
            new_string = data
            new_string = new_string.replace(" {}(.".format(first_word), " U0(.")
            for a in range(len(nets)):
                net_name = nets[a]
                port_name = mapping[first_word][0][a]
                if not behav:
                    net_name = "({0})".format(net_name)
                    port_name = "({0})".format(port_name)
                new_string = new_string.replace(port_name, net_name)
            new_hdl.append(new_string[:-1])
        else:
            new_hdl.append(line)
    return "\n".join(new_hdl)


def increment_iname(hdl):
    """Assigns unique instances names for all cells in the HDL block

    By default, instances are named U0, U1, U2, etc.
    These names need to be made unique.
    """
    U_count = 0
    good_hdl = ""
    while True:
        # Find the next instance name
        U = re.search(r"U\d+", hdl)
        if U is None:
            break
        # Replace it with the next name
        good_hdl += hdl[: U.start()] + "U" + str(U_count)
        hdl = hdl[U.end() :]
        U_count += 1
    return good_hdl + hdl


def increment_wname(hdl, w_ctr):
    """Assigns unique instances names for all internal wires in the HDL block

    By default, internal wires are named w0, w1, w2, etc.
    These names need to be made unique.
    """
    w = re.findall(r"w\d+", hdl)
    # Get the list of all wire names, in reverse
    w = sorted(list(set(w)), reverse=True)
    # Substitute them in reverse as well
    # This avoids the issue of 0 -> 1, 1 -> 2, 2 -> 3 making all wires == 3
    # Instead, 2 -> 3, 1 -> 2, 0 -> 1, making the wires good
    w_ctr = w_ctr + len(w)
    final_w_ctr = w_ctr
    for x in w:
        hdl = hdl.replace(x, "w{}".format(final_w_ctr))
        final_w_ctr -= 1
    return hdl, w_ctr


def display_png(graph, *args, **kwargs):
    """Given a graph, executes its png() method and displays the image

    Args:
        graph (ExpressionGraph): The graph to be rendered
        *args: The arguments to pass to the graph's png() method
        **kwargs: The keyword arguments to pass to the graph's png() method
    """
    # Get a temporary file name
    fname = str(uuid.uuid4()) + ".png"
    # Execute the function
    graph.png(*args, out=fname, **kwargs)
    # Get the PNG data
    ret = open(fname, "rb").read()
    # Delete the PNG file
    os.remove(fname)

    return ret


def display_gif(graphs, *args, **kwargs):
    """Given a list of graphs, executes their png() method and displays a gif

    Args:
        graphs (list of ExpressionGraph): The graphs to be rendered
        *args: The arguments to pass to the graphs' png() method
        **kwargs: The keyword arguments to pass to the graphs' png() method
    """
    # Get temporary file names
    fnames = [str(uuid.uuid4()) + ".png" for _ in range(len(graphs))]
    # Execute the function
    for graph, fname in zip(graphs, fnames):
        graph.png(*args, out=fname, **kwargs)
    # Collect the images
    images = [PIL.Image.open(fname) for fname in fnames]
    # Get maximum width and height
    max_bbox = [0, 0, 0, 0]
    for a in images:
        bbox = a.getbbox()
        if bbox[0] < max_bbox[0]:
            max_bbox[0] = bbox[0]
        if bbox[1] < max_bbox[1]:
            max_bbox[1] = bbox[1]
        if bbox[2] > max_bbox[2]:
            max_bbox[2] = bbox[2]
        if bbox[3] > max_bbox[3]:
            max_bbox[3] = bbox[3]
    for a in range(len(images)):
        im = images[a]
        im = im.crop(max_bbox)
        images[a] = im
    # Save the GIF file
    gif_name = str(uuid.uuid4()) + ".gif"
    images[0].save(
        gif_name,
        save_all=True,
        format="GIF",
        duration=1000,
        loop=0,
        disposal=2,
        transparency=255,
        append_images=images,
    )
    # Get the GIF data
    ret = open(gif_name, "rb").read()
    # Delete the PNG files
    for fname in fnames:
        os.remove(fname)
    # Delete the GIF file
    os.remove(gif_name)

    return ret


if __name__ == "__main__":
    raise RuntimeError("This file is importable, but not executable")



from .util import change_in_nets


class EquivClass:
    """Defines an equivalence class of ExpressionNodes

    This represents nodes that are mathematically distinct,
    but physically equivalent.

    Attributes:
        nodes (set of ExpressionNode): The nodes in this equivalence class
        rep (ExpressionNode): The representative of this equivalence class
        out_nets (dict): The output nets of this equivalence class
        parents (set of EquivClass): The parents of this equivalence class
    """

    def __init__(self, rep):
        """Initializes an EquivClass object

        Args:
            rep (ExpressionNode): The representative of this equivalence class
            nodes (list of ExpressionNode): The nodes in this equivalence class
        """
        self.rep = rep
        self.nodes = {rep}
        self.out_nets = rep.out_nets
        self.parents = set()

    def __len__(self):
        """Returns the number of nodes in this equivalence class"""
        return len(self.nodes)

    def __iter__(self):
        """Iterates over the nodes in this equivalence class"""
        return iter(self.nodes)

    def equiv(self, other):
        """Returns whether this equivalence class is equivalent to another

        Args:
            other (EquivClass): The equivalence class to compare to
        """
        if not isinstance(other, EquivClass):
            raise TypeError("other must be an EquivClass")

        other_rep = other.rep
        # Check whether the two representatives are equivalent
        if self.rep > other.rep or self.rep < other.rep:
            return False
        if self.rep.value != other.rep.value:
            return False

        # Check whether their subtrees are also equivalent
        for a in range(len(self.rep.children)):
            this_c = self.rep[a]
            if this_c is not None:
                try:
                    other_c = other_rep[a]
                except IndexError:
                    return False
                if not this_c.equiv_class == other_c.equiv_class:
                    return False

        return True

    def __str__(self):
        """Returns a string representation of this equivalence class"""
        return "ec({}@{})#{}".format(self.rep, self.rep.graph, len(self))

    def __repr__(self):
        return str(self)

    def _recalculate_parents(self):
        """Recalculates the parents of this equivalence class"""
        self.parents = set()
        for node in self:
            parent = node.parent
            if parent is None:
                self.parents.add(None)
            else:
                self.parents.add(parent.equiv_class.rep)
        return

    def change_rep(self, new_rep):
        """Changes the representative of this equivalence class"""
        if new_rep not in self.nodes:
            raise ValueError("new_rep must be in this equivalence class")

        for node in self:
            parent = node.parent
            index = parent.children.index(node)
            change_in_nets(parent, self.out_nets, new_rep.out_nets, index)
        self.rep = new_rep
        self.out_nets = new_rep.out_nets
        return

    def merge(self, other, check_equiv=True):
        """Merges this equivalence class with another equivalence class

        Args:
            other (EquivClass): The equivalence class to merge with
        """
        if not isinstance(other, EquivClass):
            raise TypeError("other must be an EquivClass")
        if check_equiv and not self.equiv(other):
            raise ValueError("Cannot merge non-equivalent equivalence classes")

        # Merge the equivalence classes
        self.nodes |= other.nodes
        self.parents |= other.parents

        # Merge the out_nets
        for k, v in self.out_nets.items():
            if v is [None]:
                self.out_nets[k] = other.out_nets[k]

        # Assign the correct equivalence class to new nodes
        # Overwrite input nets of new nodes
        for node in other.nodes:
            node.equiv_class = self
            parent = node.parent
            index = parent.children.index(node)
            change_in_nets(parent, node.out_nets, self.rep.out_nets, index)

        return self

    def is_bifurcation(self):
        """Returns whether this equivalence class bifurcates

        That is to say, this answers ∃n ≡ self s.t n.parent !≡ self.parent
        """
        return len(self.parents) > 1

    def reset(self):
        """Resets the equivalence classes of all nodes in this class"""
        for node in self:
            node.equiv_class = EquivClass(node)
            parent = node.parent
            index = parent.children.index(node)
            change_in_nets(parent, self.rep.out_nets, node.out_nets, index)
        return


if __name__ == "__main__":
    raise RuntimeError("This module is not intended to be run directly")



from .ExpressionGraph import ExpressionGraph
from .ExpressionTree import ExpressionTree
from .util import display_gif, increment_iname, increment_wname, wrap_quotes


class ExpressionForest(ExpressionGraph):
    """Defines a forest of binary expression trees

    Attributes:
        trees (list): A list of expression trees
        tree_type (class): The type of tree this forest contains
        width (int): The number of leaves in the forest
        radix (int): The radix of the tree
        idem (bool): Whether the trees' operators are idempotent
        node_defs (dict): A dictionary that defines the tree's nodes
        in_shape (list of int): The shape of each leaf node's input
        out_shape (list of int): The shape of the root node's output
        cocycle_shape (int, int): The shape of the main recurrence node's output
        equiv_classes (list of EquivClass): The equivalence classes of the graph

    Attributes inherited from ExpressionGraph:
        name (string): The name of the graph
        next_net (int): The next net name to be used
        next_block (int): The next block name to be used
        blocks (list): The list of blocks in the graph
        in_ports (list of ((string, int), string)): The list of input ports
        out_ports (list of ((string, int), string)): The list of output ports
    """

    # NOTE: This function fails flake8 C901
    # TO-DO: Make this function pass flake8 C901
    def __init__(
        self,
        width,
        in_ports,
        out_ports,
        tree_type=ExpressionTree,
        name="forest",
        alias=None,
        tree_start_points=None,
        initialized_trees=None,
        radix=2,
        idem=False,
        node_defs={},
    ):
        """Initializes the ExpressionForest

        There are three arguments that can be used to intialize the forest.
        This is their priority order, from highest to lowest:
            - initialized_trees
            - tree_start_points
            - alias

        Args:
            width (int): The number of leaves in the forest
            in_ports (list of ((string, int), string)): List of input ports
            out_ports (list of ((string, int), string)): List of output ports
            tree_type (class): The type of tree to use
            name (string): The name of the graph
            alias (string): The starting structure of the forest [LEGACY]
            tree_start_points (list of int): Catalan IDs for each tree
            initialized_trees (list): A list of trees to initialize the forest
                If this parameter is set,
                the forest will undergo an alternate constructor
            radix (int): The radix of the tree
            idem (bool): Whether the tree's main operator is idempotent
            node_defs (dict): A dictionary that must define these nodes:
                - "pre": Pre-processing node
                - "root": Root node
                - "cocycle": Main expression node used in the tree
                - "buffer": Buffer node

            Optional node definitions include but are not limited to:
                - "rspine": Nodes that lie along the right spine of the tree
                - "lspine": Nodes that lie along the left spine of the tree
                - "lspine_pre": Pre- node that feeds into the left spine
                - "first_pre": Right-most pre-processing node
        """
        if not isinstance(width, int):
            raise TypeError("Forest width must be an integer")
        if width < 1:
            raise ValueError("Forest width must be at least 1")
        if alias not in [
            None,
            "serial",
            "ripple",
            "ripple-carry",
            "sklansky",
            "kogge-stone",
            "brent-kung",
        ]:
            error = "Forest start point {0} is not implemented.\n"
            error += "Consider using non-legacy start points.\n"
            error += "These correspond to the trees' Catalan IDs."
            error = error.format(alias)
            raise NotImplementedError(error)
        if not isinstance(radix, int):
            raise TypeError("Tree radix must be an integer")
        if not isinstance(name, str):
            raise TypeError("Tree name must be a string")
        if not isinstance(idem, bool):
            raise TypeError("Tree idempotency must be a boolean")
        if not isinstance(node_defs, dict):
            raise TypeError("Tree node definitions must be a dictionary")
        for required in ["pre", "root", "cocycle", "buffer"]:
            if required not in node_defs:
                raise ValueError(
                    (
                        "Tree node definitions must contain" " the node {}"
                    ).format(required)
                )

        # If both kinds of start points are specified,
        # Use the non-legacy kind
        if alias is not None and tree_start_points is not None:
            alias = None

        # Save constructor arguments
        self.width = width
        self.tree_type = tree_type
        self.radix = radix
        self.idem = idem
        self.node_defs = node_defs

        # If initialized trees are provided, simply use them
        if isinstance(initialized_trees, list):
            if len(initialized_trees) != width:
                raise ValueError("Number of trees in a forest must equal width")
            self.trees = initialized_trees
            super().__init__(name=name, in_ports=in_ports, out_ports=out_ports)
            return

        # Otherwise, initialize the trees
        self.trees = []

        for a in range(1, width + 1):
            ### NOTE: INPUT SHAPE IS CURRENTLY ASSUMED TO BE [1,1,1,..]
            ### TO-DO: Allow for arbitrary input shape
            ### Need to check whole file for this implicit assumption
            if a == 1:
                bit_s = "[{0}]".format(a - 1)
            else:
                bit_s = "[{0}:0]".format(a - 1)
            tree_in_ports = [((x[0][0], a), x[1] + bit_s) for x in in_ports]

            ### NOTE: TREE OUTPUT SHAPE IS CURRENTLY ASSUMED TO BE [1,1,1,..]
            ### TO-DO: Allow for arbitrary input shape
            ### Need to check whole file for this implicit assumption
            bit_s = "[{0}]".format(a - 1)
            tree_out_ports = [((x[0][0], 1), x[1] + bit_s) for x in out_ports]

            if tree_start_points is not None and len(tree_start_points) > a - 1:
                catalan_id = tree_start_points[a - 1]
            else:
                catalan_id = 0

            t = self.tree_type(
                width=a,
                in_ports=tree_in_ports,
                out_ports=tree_out_ports,
                name="tree_{}".format(a - 1),
                radix=radix,
                start_point=catalan_id,
            )
            self.trees.append(t)

        # Initialize the graph
        super().__init__(name=name, in_ports=in_ports, out_ports=out_ports)
        self.equiv_classes = set()

        # If tree_start_points was provided, ignore the alias
        if tree_start_points is not None:
            return

        # Transform the forest towards the starting point
        if alias in ["serial", "ripple", "ripple-carry"]:
            pass
        elif alias == "sklansky":
            for t in self.trees[2:]:
                t.rbalance(t.root[1])
        elif alias == "kogge-stone":
            for t in self.trees[2:]:
                t.lbalance(t.root[1])
                t.equalize_depths(t.root[1])
        elif alias == "brent-kung":
            for t in self.trees[2:]:
                t.rbalance(t.root[1])
                while not t.root[1][0].is_proper():
                    t.right_rotate(t.root[1][0])
            self.calculate_fanout()
            self.decouple_all_fanout()

    def __getitem__(self, key):
        """Returns the tree at the given index

        Args:
            key (int): The index of the tree to return

        Returns:
            ExpressionTree: The tree at the given index
        """
        return self.trees[key]

    def __iter__(self):
        """Iterates over the trees in the forest"""
        return iter(self.trees)

    def _repr_png_(self):
        """Automatically display diagrams in a Notebook"""
        return display_gif(self.trees)

    ### NOTE: Equivalence requires nodes to be fully equivalent
    ### It is possible for two nodes to be partially equivalent
    ### For example, if their subtrees are identical
    ### But one node's expression is a subset of the other
    ### Such as (+,-) vs (+) or (-)
    ### This should be accounted for
    def find_equivalent_nodes(self):
        """Finds equivalent nodes amongst the forest's trees"""
        for i1 in reversed(range(len(self.trees))):
            t1 = self.trees[i1]
            for i2 in range(i1):
                t2 = self.trees[i2]
                for n1 in t1.nodes:
                    for n2 in t2.nodes:
                        # Try to merge equivalence classes
                        # Raise exception if they are not equivalent
                        try:
                            old_ec = n2.equiv_class
                            n1.equiv_class.merge(n2.equiv_class)
                            self.equiv_classes.add(n1.equiv_class)
                            if old_ec in self.equiv_classes:
                                self.equiv_classes.remove(old_ec)
                        except ValueError:
                            pass
        for ec in self.equiv_classes:
            ec._recalculate_parents()
        self.mark_equivalent_nodes()

    def mark_equivalent_nodes(self):
        """Mark all redundant nodes with stripes on diagrams"""
        for t in self.trees:
            for n in t.nodes:
                if n.equiv_class.rep is not n:
                    if t.nodes(data=True)[n].get("gradientangle", "0") == "135":
                        continue
                    col = t.nodes(data=True)[n]["fillcolor"]
                    t.nodes(data=True)[n]["orig_color"] = col
                    new_col = wrap_quotes("red;0.5:{0};0.5".format(col))
                    t.nodes(data=True)[n]["fillcolor"] = new_col
                    t.nodes(data=True)[n]["gradientangle"] = "135"

    def reset_equivalent_nodes(self):
        """Resets the equivalence classes of all nodes"""
        self.unmark_equivalent_nodes()
        for ec in self.equiv_classes:
            ec.reset()

    def unmark_equivalent_nodes(self):
        """Remove stripes from all redundant nodes on diagrams"""
        for t in self.trees:
            for n in t.nodes:
                if t.nodes(data=True)[n].get("gradientangle", "0") == "135":
                    col = t.nodes(data=True)[n]["fillcolor"]
                    old_col = t.nodes(data=True)[n].get("orig_color", col)
                    t.nodes(data=True)[n]["fillcolor"] = old_col
                    t.nodes(data=True)[n]["gradientangle"] = "0"

    ### NOTE: Where this logic belongs is an open question
    def find_parallel_tracks(self):
        """Finds parallel tracks amongst the forest's trees"""
        for i1 in reversed(range(len(self.trees))):
            t1 = self.trees[i1]
            for i2 in range(i1):
                t2 = self.trees[i2]
                for n1 in t1.nodes:
                    for n2 in t2.nodes:
                        if n1.tracks(n2):
                            n1.set_tracks(n2)

    ### NOTE: Improve fanout estimate
    def _calc_node_fanout(self, node):
        """Estimates the delay caused by fanout for a node

        A node has fanout of k if:
         - node is the representative of an equivalence class
         - There are k nodes in the forest such that
             - They belong to node's equivalence class
             - Their parents belong to unique equivalence classes

        For example, if an equivalence class contains the nodes
        n1, n2, n3, n4; with n1 being the representative
        And there is an equivalence relation between n1.parent and n2.parent
        but not between any other parents

        n1 has a fan-out of 3
        The other nodes do not matter
        """
        tree = node.graph
        # Grab the node's equivalence class
        ec = node.equiv_class
        # If node is not ec's representative, do nothing
        if ec.rep is not node:
            return
        # If node is root, do noting
        if node.parent is None:
            return

        # Count the fanout
        fanout = len(ec.parents)

        # Modify the relevant edge's data
        e_data = tree.get_edge_data(node.parent, node)
        index = node.parent.children.index(node)
        g = node.node_data["le"][index]
        ### This is a bad estimate of fanout's effect on delay
        e_data["fanout"] = g * fanout
        tree.update_edge_weight(node.parent, node)

        return fanout

    def calculate_fanout(self):
        """Calculates the fanout of all nodes in the forest

        See the description of _calc_node_fanout for details
        """
        for t in self.trees:
            for n in t:
                self._calc_node_fanout(n)

    ### NOTE: Improve cross-coupling capacitance estimate
    def _calc_node_tracks(self, node):
        """Estimates the delay caused by parallel wires for a node

        A node has tracks of k if:
         - node is the representative of an equivalence class
         - There are k nodes in the forest such that
            - They are the heads of their own equivalence classes
            - node.y_pos = other.y_pos
            - node < other, other < node.parent, node.parent < other.parent
            or
            - node > other, node < other.parent, other.parent > node.parent
        """
        tree = node.graph
        # If node is not its equivalence class representative, do nothing
        if node.equiv_class.rep is not node:
            return
        # If node is root, do noting
        if node.parent is None:
            return

        # Compare node to all other nodes in the forest
        for t in self.trees:
            for n in t:
                if n.equiv_class.rep is not n:
                    continue
                if node.tracks(n):
                    node.set_tracks(n)

        # Count the tracks
        tracks = len(node.tracks_class)

        # Modify the relevant edge's data
        e_data = tree.get_edge_data(node.parent, node)
        ### This is a bad estimate of tracks' effect on delay
        e_data["tracks"] = tracks
        tree.update_edge_weight(node.parent, node)

        return tracks

    def calculate_tracks(self):
        """Calculates the tracks of all nodes in the forest

        See the description of _calc_node_tracks for details
        """
        for t in self.trees:
            for n in t:
                self._calc_node_tracks(n)

    def decouple_fanout(self, node, tree):
        """Decouples the fanout of the specified node in the specified tree"""

        # Check if arguments are valid
        if not (node in node.graph):
            raise ValueError("Node is not in its graph. Impossible error.")
        if not (node.graph in self.trees):
            raise ValueError(
                "Node graph is not in this forest. Impossible error."
            )
        if not (tree in self.trees):
            raise ValueError("Trying to decouple fanout in an invalid tree.")

        # Look inside the node's equivalence class for an other_node from tree
        other_node = None
        ec = node.equiv_class
        for a in ec:
            if a.graph is tree:
                other_node = a
                break
        if other_node is None:
            raise ValueError("Trying to decouple fanout in innapropriate tree.")

        # Decouple fanout
        buffer = tree.insert_buffer(other_node)

        # Fix equivalence classes
        for p in ec.parents:
            # Try to add new buffer to other equivalence classes
            try:
                p.merge(buffer.equiv_class)
            except ValueError:
                pass
        ec.parents.add(buffer)

        # Fix fanout
        for a in ec:
            self._calc_node_fanout(a)
        for a in buffer.equiv_class:
            self._calc_node_fanout(a)

    def optimize_nodes(self):
        """Greedily attempt to swap in nodes with same footprint

        All node node_data have a footprint attribute clarifying which node_data
        refer to the same node concept. All node node_data also have a priority
        attribute, used to determine which module is most "optimal".

        This method attempts to raise the total optimality of the forest by
        swapping in higher-priority node_data.

        This can probably be safely executed at any point in time, but it is
        advisable to only execute this once no more rotations will take place.

        Args:
            node (Node): The root node of the sub-tree to optimize
        """
        for t in self.trees:
            t.optimize_nodes()

    def add_best_blocks(self, graph_type=None):
        """Calls upon the Trees' add_best_blocks methods"""
        for t in self.trees:
            t.add_best_blocks(graph_type=graph_type)

    def hdl(
        self,
        out=None,
        optimization=1,
        mapping="behavioral",
        language="verilog",
        flat=False,
        block_flat=False,
        node_flat=True,
        merge_mapping=True,
        module_name=None,
        description_string="start of unnamed graph",
    ):
        """Creates a HDL description of the forest

        Args:
            out (str): The file to write the HDL to
            optimization (int): The optimizaton level to use
                optimization = 0: No optimization
                optimization = 1: Remove some redundant logic out of nodes
                optimization = 2: Perform previous optimization, then partition
                the HDL into smaller blocks which synthesis tools can handle
                better.
            mapping (str): The technology mapping to use
            language (str): The language in which to generate the HDL
            flat (bool): If True, flatten the graph's HDL
            block_flat (bool): If True, flatten all blocks in the graph's HDL
            node_flat (bool): If True, flatten all cells in the graph's HDL
            merge_mapping (bool): If True, merge the mapping file in
            module_name (str): The name of the module to generate
            description_string (str): String commend to prepend to the HDL

        Returns:
            str: HDL module definition representing the graph
            list: Set of HDL module definitions used in the graph

        """
        # If module name is not defined, set it to graph's name
        if module_name is None:
            module_name = self.name

        if optimization > 0:
            self.optimize_nodes()
        if optimization > 1:
            self.find_equivalent_nodes()
            self.calculate_fanout()
            self.calculate_tracks()
            self.add_best_blocks()

        hdl = []
        module_defs = set()
        # This HDL description will have multiple cell-internal wires
        # By default, these are named w*
        # These names need to be made unique
        w_ctr = 0
        for t in reversed(self.trees):
            desc = "{}_forest {}".format(self.name, t.name)
            t_hdl, t_module_defs = t.hdl(
                language=language,
                mapping=mapping,
                flat=flat,
                block_flat=block_flat,
                node_flat=node_flat,
                merge_mapping=merge_mapping,
                module_name="{0}_{1}".format(module_name, t.name),
                description_string=desc,
            )
            t_hdl, w_ctr = increment_wname(t_hdl, w_ctr)
            hdl.append(t_hdl)
            module_defs |= t_module_defs
        hdl = "\n".join(hdl)

        # This HDL description will have multiple instances in it
        # By default, util.hdl_inst names all instances "U0"
        # These names need to be made unique
        hdl = increment_iname(hdl)

        hdl, module_defs, file_out_hdl = self._wrap_hdl(
            hdl, module_defs, language, module_name
        )
        if out is not None:
            self._write_hdl(file_out_hdl, out, language, mapping, merge_mapping)

        return hdl, module_defs, file_out_hdl

    def gif(self, out="forest.gif"):
        with open(out, "wb") as fout:
            fout.write(self._repr_png_())

    ### NOTE: ALL METHODS BELOW ARE FOR LEGACY SUPPORT ONLY ###

    def LF(self, x, y=None):
        """Performs an LF transform on the specified node in all trees

        This is node by calling on the corresponding method in the Tree class.
        See the disclaimer inside the docstring of the ExpressionTree method.
        """

        # Try to first apply transform on most significant tree
        # Once a valid transform is applied, use x,y for all other trees
        for t in reversed(list(self.trees)):
            if x < t.width - 1:
                attempt = t.LF(x, y)
                if attempt is not None:
                    x, y = attempt
        return x, y

    def FL(self, x, y=None):
        """Performs an FL transform on the specified node in all trees

        This is node by calling on the corresponding method in the Tree class.
        See the disclaimer inside the docstring of the ExpressionTree method.
        """

        # Try to first apply transform on most significant tree
        # Once a valid transform is applied, use x,y for all other trees
        for t in reversed(list(self.trees)):
            if x < t.width - 1:
                attempt = t.FL(x, y)
                if attempt is not None:
                    x, y = attempt
        return x, y

    def TF(self, x, y=None):
        """Performs a TF transform on the specified node in all trees

        This is node by calling on the corresponding method in the Tree class.
        See the disclaimer inside the docstring of the ExpressionTree method.
        """

        # Try to first apply transform on most significant tree
        # Once a valid transform is applied, use x,y for all other trees
        for t in reversed(list(self.trees)):
            if x < t.width - 1:
                attempt = t.TF(x, y)
                if attempt is not None:
                    x, y = attempt
        return x, y

    def FT(self, x, y=None):
        """Performs an FT transform on the specified node in all trees

        This is node by calling on the corresponding method in the Tree class.
        See the disclaimer inside the docstring of the ExpressionTree method.
        """

        # Try to first apply transform on most significant tree
        # Once a valid transform is applied, use x,y for all other trees
        for t in reversed(list(self.trees)):
            if x < t.width - 1:
                attempt = t.FT(x, y)
                if attempt is not None:
                    x, y = attempt
        return x, y

    def legacy_decouple_node_fanout(self, node, maximum_fanout=2):
        """Decouples the fanout of the specified node in all trees"""

        # Get node's equivalence class
        ec = node.equiv_class
        # Filter out nodes whose parents are equivalent
        ec = [
            x
            for x in ec
            if x.parent is None or x.parent.equiv_class.rep is x.parent
        ]
        # Decouple fanout
        ctr = 0
        for a in range(len(ec)):
            n = ec[a]
            parent = n.parent
            if parent is None:
                continue
            # The last connection does not benefit from extra buffer
            if a == len(ec) - 1:
                ctr -= 1
            index = parent.children.index(n)
            for b in range(ctr // (maximum_fanout - 1)):
                for p in parent.equiv_class:
                    t = p.graph
                    t.insert_buffer(p[index])
            ctr += 1

    def decouple_all_fanout(self, maximum_fanout=2):
        """Decouples the fanout of all nodes in all trees"""
        targets = []
        for t in reversed(self.trees[1:]):
            for n in t:
                if (n.parent is not None) and (n.equiv_class.rep is n):
                    targets.append(n)
        for n in targets:
            self.legacy_decouple_node_fanout(n, maximum_fanout)

    ### NOTE: END LEGACY METHODS ###


if __name__ == "__main__":
    raise RuntimeError("This file is importable, but not executable")



name = "ppa_post_nolspine"
data = dict()

### Post-processing node
data[
    "verilog"
] = """
module ppa_post_nolspine(xin, gin, sum);

	input xin, gin;
	output sum;

	xor2 U1(sum,gin,xin);

endmodule
"""

data[
    "vhdl"
] = """
entity ppa_post_nolspine is
	port (
		xin : in std_logic;
		gin : in std_logic;
		sum : out std_logic
	);
end entity;

architecture behavior of ppa_post_nolspine is
begin

U1: xor2
	port map (
		A => xin,
		B => gin,
		Y => sum
	);

end architecture;
"""

data["shape"] = "circle"
data["label"] = "x"
data["fontname"] = "Comic Sans"
data["style"] = "bold"
data["fixedsize"] = "shape"
data["penwidth"] = "4.0"
data["fontsize"] = "52"

data["ins"] = [("gin", 1, 0, 1), ("xin", 1, 1, 0)]
data["outs"] = [("sum", 1)]

data["logic"] = lambda xin, gin: [xin ^ gin]

data["pd"] = 9 / 3
data["le"] = [9 / 3, 9 / 3]



from importlib import import_module
from pathlib import Path

for f in Path(__file__).parent.glob("*.py"):
    module_name = f.stem
    if (not module_name.startswith("_")) and (module_name not in globals()):
        import_module(f".{module_name}", __package__)
    del f, module_name
del import_module, Path

__all__ = ["node_data"]

imported = [globals()[x] for x in dir() if x[0] != "_"]
node_data = {}
for x in imported:
    node_data[getattr(x, "name")] = getattr(x, "data")



name = "ppa_lspine_single"
data = dict()

### ppa_lspine_single
data[
    "verilog"
] = """
module ppa_lspine_single(xin, yin, pin, gin, xout, yout);

	input xin, yin, pin, gin;
	output xout, yout;

	mux2 U2(xout,gin,xin,yin);
	mux2 U3(yout,pin,xin,yin);

endmodule
"""

data[
    "vhdl"
] = """
"""

data["shape"] = "circle"
data["label"] = "x"
data["fontname"] = "Comic Sans"
data["style"] = "bold"
data["fixedsize"] = "shape"
data["penwidth"] = "4.0"
data["fontsize"] = "52"

data["ins"] = [
    ("xin", 1, 1, 0),
    ("yin", 1, 1, 0),
    ("pin", 1, 0, 1),
    ("gin", 1, 0, 1),
]
data["outs"] = [("xout", 1), ("yout", 1)]

# ppa_lspine_single["logic"] = lambda pin, gin: [pin ^ gin, ~(pin ^ gin)]

data["pd"] = 9 / 3
data["le"] = [9 / 3, 9 / 3]
